unit Alternation;

{
  Essential for Particle system, it describes how Particle changes (or what do that Particle)
    - BoolFunction -> says when apply some change
    - Selector     -> says to which Particle apply this change (if nil then to receiver)
    - Change       -> what will be done to Particle
                      it gets information about who wants to change whom and how much time passed
}
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, DeletionList, BoolFunction, Selector, Change;

type
  DL_Change = specialize TDeletionList<TChange>;

  //this class is not for inheriting, it is the base of ParticleSystem
  TAlternation = class
    public
      constructor Create(
        bool_fn : TBoolFunction;//nil bool_fn  -> TBoolTrue will be alocated
        sele_cl : TSelector;    //nil selector -> sender will be used
        ch1     : TChange;
        ch2     : TChange = nil;
        ch3     : TChange = nil;
        ch4     : TChange = nil;
        ch5     : TChange = nil
      );
      destructor Destroy; override;

      //if Selector is nil, then sender will be changed
      //otherwise particle from Selector will be changed and Change can use data from sender
      procedure Execute(sender : TObject = nil; time_diff : Real = 0);

      //for deletion list
      procedure Evaluate;        //updates changes
      function IsDead : Boolean; //when I`m not neccesary any more

    private
      //in human language : 'Use data from sender to change Receiver'
      //or when Selector is nil then receiver will be the caster
      // (the Particle passed in Group.Update and Particle.Evaluate)
      procedure ApplyTo(receiver : TObject; sender : TObject; time_diff : Real = 0);

      bfn : TBoolFunction;
      sel : TSelector;
      dlCh : DL_Change;     //it can have more changes, sometimes its useful
  end;

implementation

//if nil selector, then sender is sended to Change
constructor TAlternation.Create(
        bool_fn : TBoolFunction;
        sele_cl : TSelector;
        ch1     : TChange;
        ch2     : TChange = nil;
        ch3     : TChange = nil;
        ch4     : TChange = nil;
        ch5     : TChange = nil
);
begin
  //Writeln('TAlternation.Create');

  if bool_fn <> nil then bfn := bool_fn
  else bfn := TBoolTrue.Create;  //default (always true)

  sel := sele_cl;

  if (ch1 <> nil) or (ch2 <> nil) or (ch3 <> nil) or (ch4 <> nil) or (ch5 <> nil) then begin
    dlCh := DL_Change.Create;
    if ch1 <> nil then dlCh.Add(ch1);
    if ch2 <> nil then dlCh.Add(ch2);
    if ch3 <> nil then dlCh.Add(ch3);
    if ch4 <> nil then dlCh.Add(ch4);
    if ch5 <> nil then dlCh.Add(ch5);
  end;
end;

destructor TAlternation.Destroy;
begin
  if bfn <> nil then bfn.Free;
  if sel <> nil then sel.Free;
  if dlCh <> nil then dlCh.Free;
  inherited;
  //Writeln('TAlternation.Destroy');
end;

procedure TAlternation.Execute(sender : TObject; time_diff : Real);
var
  p : pointer;
  b : Boolean;
begin
  if (bfn = nil) or (dlCh = nil) then exit;
  if (sender = nil) and (sel = nil) then exit;

  b := bfn.Value(Sender);
  if not b then exit; //nothing will be done

  //if no Selector
  if sel = nil then
    begin
      //who sends it, he wants to be changed (so he will be the receiver in ApplyTo)
      ApplyTo(sender, nil, time_diff);
    end
  //if there is some selector
  else
  begin
    //fetch all particles and call on them ApplyTo( foreach element in Selected)
    sel.Select(sender);
    p := sel.first;
    while p <> nil do begin
      //the receiver = sel.getElement(p)

      {if(sel.getElement(p) = sender) then
        TBoolFunction(p).Value(sender);}

      if(sel.getElement(p) <> sender) then
        if((sel.pair_fn = nil) or (sel.pair_fn.Value(sel.getElement(p), sender))) then
          ApplyTo(sel.getElement(p), sender, time_diff);

      sel.next(p);
    end;
    sel.Deselect;
  end;
end;

procedure TAlternation.Evaluate;
begin
  dlCh.Update;
  if bfn <> nil then bfn.Evaluate;
  if sel <> nil then sel.Evaluate;
end;

function TAlternation.IsDead : Boolean;
begin
  //I`m broken
  Result := (bfn = nil) or (dlCh = nil);
  if Result then exit;

  //nothing to do
  Result := (dlCh.Empty) or (bfn.IsDead);
  if Result then exit;

  //if selector (if exists) says that nothing will be selected all the time
  if sel <> nil then Result := sel.IsDead;
end;

procedure TAlternation.ApplyTo(receiver, sender : TObject; time_diff : Real);
var
  iter : pointer;
begin

 //to all Change apply this Alternation
  iter := dlCh.first;
  while iter <> nil do begin
    (dlCh.getElement(iter)).Apply(receiver, sender, time_diff);
    dlCh.next(iter);
  end;
end;

end.

