unit ParticleChange;

{
  Set of Classes which Change Particles, necessary for Alternation

  !Apply have TObject parameters, so they must be retyped, for more info about TChange see Change.pas
}
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Change, Particle, Vektor, Graphics, ParticlePairBoolFunction;

type

  //Moves Particle by its velocity
  TChangeMove = class(TChange)
    public
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
  end;
  TChangeSpeedLimiter = class(TChange)
    public
      //friction should be between 0, 1 (0 -no friction, 1- full friction I will stop at once)
      //also the area of shape is considered
      constructor Create(speed_limit : Real = -1.0; friction : Real = 0.0; min_velocity : Real = 0.0);
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
      sp_lim : Real;
      fr     : Real;
      min_vel: Real;
  end;

  TChangeStayInBox= class(TChange)
    public
      constructor Create(x1, y1, x2, y2 : Real);
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
    //describe a box
      a,b : Real;  //a = min(a, c), b = min(b,d)
      c,d : Real;  //c = max(a, c), d = max(b,d)
  end;

  TChangeCollision= class(TChange)
    public
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
  end;

  TChangeGravity= class(TChange)
    public
      // power = kappa, minimal distance to affect (prevents div by zero)
      // positive -> I will here object
      // positive = false (negative) i will repel objects
      constructor Create(
        power : Real;
        minimal_distance : Real = 1.0;
        position : TVektor = nil;
        positive : Boolean = true
      );
      destructor Destroy; override;
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
      kappa : Real;
      min_dist : Real;
      direction : Boolean;
      pos : TVektor;
  end;

  TChangeDamage= class(TChange)
    public
      constructor Create(x : Integer; hurt_rec:Boolean=true; hurt_sen : Boolean=false); //how much damage do you want to do
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
      power : Integer;
      r, s : Boolean;
  end;

//who will be killed on apply
  TChangeKill= class(TChange)
    public
      constructor Create(kill_receiver : Boolean = true; kill_sender : Boolean = 0);
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
      kill_rec, kill_sen : Boolean;
  end;

//the difference vektor will be scaled by this, if scale = 0.0 then there will be static relation
//remark, negative acc_scale will make particles distract
//you can specify who will be followed (auto Dead management)
  TChangeFollowParticle= class(TChange)
    public
      constructor Create(); overload;
      constructor Create(position_diff : TVektor; acc_scale : Real = 0.0; particle_follow : TParticle = nil); overload;
      destructor Destroy; override;
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;

      //par_follow.IsDead management
      procedure Evaluate; override;

      par_follow : TParticle; //by default (par_follow = nil) is it parent
    protected
      pos_diff : TVektor;
      scale : Real;
  end;

// TChangeEmit= class(TChange) -> in ParticleEmitor

  TChangeColorDependOnHealth= class(TChange)
    public
      constructor Create(red1, green1, blue1,
                         red2, green2, blue2 : Integer;
                         from : Integer = 100; til : Integer = 0);
      procedure Apply(receiver:TObject; sender:TObject=nil; time_diff : Real = 0); override;
    protected
      r1,g1,b1 : Integer;  //from what color
      r2,g2,b2 : Integer;  //to color
      a1    : Integer;     //from health (default = 100)
      a2    : Integer;     //to health (default = 0)
      //formula : col = (col2 - col1) * (health - a2) / (a1 - a2)
  end;

implementation
procedure ParticleMoveBy(par : TParticle; vek : TVektor; time_diff : Real);
begin
  if not (static in par.flags) then begin
    par.pos.move(vek, time_diff);    //move scaled
  end;
end;

// ------------------------- TChangeMove -------------------

procedure TChangeMove.Apply(receiver, sender: TObject; time_diff : Real);
begin
  ParticleMoveBy(TParticle(receiver), TParticle(receiver).vel, time_diff);
end;

//-------------------------- TChangeSpeedLimiter -----------------------

constructor TChangeSpeedLimiter.Create(speed_limit : Real; friction : Real; min_velocity : Real = 0.0);
begin
  sp_lim := speed_limit;
  fr := friction;
  min_vel := min_velocity;
end;
procedure TChangeSpeedLimiter.Apply(receiver:TObject; sender:TObject; time_diff : Real);
var
  par : TParticle;
  x : Real;
begin
  par := TParticle(receiver);
  x := par.vel.size;
  if ((x > sp_lim) and (x >= 0.0)) then par.vel.scale_to(sp_lim);
  if (x < min_vel) then begin
    par.vel.x := 0;
    par.vel.y := 0;
  end;

  if (fr > 0.0) and (fr <= 1.0) then begin
    x := 1.0 - (fr * time_diff);
    par.vel.scale(x);
  end;
end;

// ------------------------- TChangeStayInBox -------------------

constructor TChangeStayInBox.Create(x1, y1, x2, y2 : Real);
begin
  a := x1;
  b := y1;
  c := x2;
  d := y2;
end;
procedure TChangeStayInBox.Apply(receiver:TObject; sender:TObject; time_diff : Real);
var
  P : TParticle;
begin
  P := TParticle(receiver);

  //Min
  if a > P.pos.x then begin
    P.pos.x := a + a - P.pos.x;
    P.vel.x := -P.vel.x;
  end;
  if b > P.pos.y then begin
    P.pos.y := a + a - P.pos.y;
    P.vel.y := -P.vel.y;
  end;

  //Max !!copy paste!!
  if c < P.pos.x then begin
    P.pos.x := c + c - P.pos.x;
    P.vel.x := -P.vel.x;
  end;
  if d < P.pos.y then begin
    P.pos.y := d + d - P.pos.y;
    P.vel.y := -P.vel.y;
  end;

end;

// -------------------------  TChangeCollision -------------------

procedure MoveParticlesTillNoCollision(P1 : TParticle; P2 : TParticle);
// !! TO DO !! -> develop collisions better !
var
  is_col : TBoolCollision;
  i : Integer;
begin
  is_col := TBoolCollision.Create;
  i:=0;
  while is_col.Value(TObject(P1), TObject(P2)) and (i < 100) do begin
    P1.pos.move(P1.vel, 0.01);
    P2.pos.move(P2.vel, 0.01);
    Inc(i,1);
  end;

  is_col.Free;
end;
procedure  TChangeCollision.Apply(
  receiver:TObject;
  sender:TObject;
  time_diff : Real = 0
);
var
  V_21, V_22 : TVektor;  //velocity vektors
  P1, P2: TParticle;
begin
  P1 := TParticle(receiver);
  P2 := TParticle(sender);

  V_21 := TVektor.Create(P2.vel);
  V_22 := TVektor.Create(P1.vel);

  //!! TO DO !!!

  P1.vel.Free;
  P1.vel := V_21;
  P2.vel.Free;
  P2.vel := V_22;

  MoveParticlesTillNoCollision(P1, P2);

end;

//------------------------------- TChangeGravity --------------------------
constructor TChangeGravity.Create(power : Real; minimal_distance : Real;
position : TVektor; positive : Boolean);
begin
  kappa := power;
  min_dist := minimal_distance;
  pos := position;
  direction := positive;
end;
destructor  TChangeGravity.Destroy;
begin
  if pos <> nil then pos.Free;
end;
procedure TChangeGravity.Apply(receiver:TObject; sender:TObject; time_diff : Real);
var
  V1, V2, A : TVektor;
  dist : Real;
begin
  if pos = nil then V1 := TParticle(sender).pos
  else V1 := pos;

  V2 := TParticle(receiver).pos;

  dist := V1.dist(V2);

  //prevents division by zero
  if dist < min_dist then dist := min_dist;

  A := TVektor.Create(V1);
  A.move(-V2.x, -V2.y);
  A.scale_to(1.0); //so I have the direction on 1 disk

  A.scale(kappa * (TParticle(sender).weight / (dist * dist)) * time_diff);
  if(not direction) then A.scale(-1.0);

  TParticle(receiver).vel.move(A);
end;

//------------------------------- TChangeDamage --------------------------
constructor TChangeDamage.Create(x : Integer; hurt_rec:Boolean=true; hurt_sen : Boolean=false);
begin
  power := x;
  r := hurt_rec;
  s := hurt_sen;
end;
procedure TChangeDamage.Apply(receiver:TObject; sender:TObject; time_diff : Real);
begin
  if(r) then TParticle(receiver).HurtMe(power);
  if(s) then TParticle(sender).HurtMe(power);
end;

//------------------------------- TChangeDamage --------------------------

constructor TChangeKill.Create(kill_receiver, kill_sender : Boolean);
begin
  kill_rec := kill_receiver;
  kill_sen := kill_sender;
end;
procedure TChangeKill.Apply(receiver:TObject; sender:TObject; time_diff : Real);
begin
  if kill_rec then TParticle(receiver).KillMe;
  if kill_sen then TParticle(sender).KillMe;
end;

//------------------------------- TChangeFollowParticle ---------------------
constructor TChangeFollowParticle.Create();
begin
  scale := 0.0;
  pos_diff := TVektor.Create(0.0, 0.0);
  par_follow := nil;
end;
constructor TChangeFollowParticle.Create(position_diff : TVektor; acc_scale : Real; particle_follow : TParticle);
begin
  scale := acc_scale;
  pos_diff := position_diff;
  par_follow := particle_follow;
end;
destructor  TChangeFollowParticle.Destroy;
begin
  if pos_diff <> nil then pos_diff.Free;
end;
procedure   TChangeFollowParticle.Apply(receiver:TObject; sender:TObject; time_diff : Real);
var
  vek : TVektor;
  parent : TParticle;
begin
  if par_follow <> nil then parent := par_follow
  else parent := TParticle(receiver).parent;

  if parent <> nil then begin
    if scale = 0.0 then begin //just translation
      //vek := TVektor.Create(TParticle(receiver).pos); //act as buffer
      TParticle(receiver).pos.copy(parent.pos);
      TParticle(receiver).pos.move(pos_diff);
    end
    else begin
     vek := TVektor.Create(parent.pos);
     vek.move(pos_diff);
     vek.move(TParticle(receiver).pos, -1);
     vek.scale(time_diff * scale);
     TParticle(receiver).vel.move(vek);
    end;
  end;
end;
procedure TChangeFollowParticle.Evaluate;
begin
  inherited;
  if par_follow <> nil then
    if par_follow.IsDead then par_follow := nil;
end;

//--------------------------------TChangeColorDependOnHealth --------------

constructor TChangeColorDependOnHealth.Create(red1, green1, blue1,
                         red2, green2, blue2,
                         from, til : Integer);
begin
  r1 := red1;
  g1 := green1;
  b1 := blue1;
  r2 := red2;
  g2 := green2;
  b2 := blue2;
  a1 := from;
  a2 := til;
end;
procedure TChangeColorDependOnHealth.Apply(receiver:TObject; sender:TObject; time_diff : Real);
var
  par : TParticle;
begin
  par := TParticle(receiver);

  if ((par.pattern <> nil) and (a1 <> a2)) then begin
    par.pattern.inards := RgbToColor(
      r2 + ( ((r1 - r2) * (par.health - a2)) div (a1 - a2)),
      g2 + ( ((g1 - g2) * (par.health - a2)) div (a1 - a2)),
      b2 + ( ((b1 - b2) * (par.health - a2)) div (a1 - a2))
    );
    //par.pattern.inards := Random(256) * 65536 + 100*256 + 10;
    //par.pattern.inards := Random(256*256*256);
  end;
end;


end.

