unit Pattern;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics;

type
  TPattern = class
    public
      constructor Create;
      destructor Destroy; override;

      //sender is who wants to draw me (a Particle usually)
      procedure Draw(img : TBitmap; x, y : Integer); virtual;

      //return some unique ID in TPattern class system
      function Id : Integer; virtual;

      //returns area of my shape
      function Area : Real; virtual;

    // --- variables ---- (in Particle is all public)
      border : Integer;   //negative border will copy color of inards
      inards : Integer;
    protected
      procedure ChangeBrushAndPen(img : TBitmap);
  end;

  type
  TPatternPoint = class(TPattern)
    public
      //sender is who wants to draw me (a Particle usually)
      procedure Draw(img : TBitmap; x, y : Integer); override;

      //return some unique ID in TPattern class system
      function Id : Integer; override;

      //returns area of my shape
      function Area : Real; override;
  end;

  TPatternCircle = class(TPattern)
    public
      constructor Create(radius : Integer);

      //sender is who wants to draw me (a Particle usually)
      procedure Draw(img : TBitmap; x, y : Integer); override;

      //return some unique ID in TPattern class system
      function Id : Integer; override;

      //returns area of my shape
      function Area : Real; override;

      r : Integer; //radius
  end;

  TPatternRectangle = class(TPattern)
    public
      constructor Create(width, height : Integer); //median will be on particle pos

      //sender is who wants to draw me (a Particle usually)
      procedure Draw(img : TBitmap; x, y : Integer); override;

      //return some unique ID in TPattern class system
      function Id : Integer; override;

      //returns area of my shape
      function Area : Real; override;

      function IsIn(x1, y1, x2, y2 : Real) : Boolean;   overload;
      function IsIn(x1, y1, x2, y2 : Integer) : Boolean;overload;
      function left : Integer;
      function right : Integer;
      function top : Integer;
      function bottom : Integer;

      w, h : Integer; //total width and height of rect (median will be in w/2, h/2)
  end;

implementation
  constructor TPattern.Create;
  begin
    border := 0;
    inards := 0;
  end;

  destructor TPattern.Destroy;
  begin
    inherited;
  end;

  procedure TPattern.Draw(img : TBitmap; x, y : Integer);
  begin

  end;

  function TPattern.Id : Integer;
  begin
    Result := -1;
  end;

function TPattern.Area : Real;
begin
  Result := 0;
end;

procedure TPattern.ChangeBrushAndPen(img : TBitmap);
begin
  img.Canvas.Brush.Color := inards;
  if border < 0 then img.Canvas.Pen.Color := inards
                else img.Canvas.Pen.Color := border;
end;

// -------------------------- TPatternPoint -------------------
procedure TPatternPoint.Draw(img : TBitmap; x, y : Integer);
begin
  ChangeBrushAndPen(img);

  img.Canvas.Rectangle(
    Round(x)-2,
    Round(y)-2,
    Round(x)+2,
    Round(y)+2
  );
end;

function TPatternPoint.Id : Integer;
begin
  Result := 1;
end;

function TPatternPoint.Area : Real;
begin
  Result := 1.0;
end;

// -------------------------- TPatternCircle -------------------
constructor TPatternCircle.Create(radius : Integer);
begin
  r := radius;
end;

procedure TPatternCircle.Draw(img : TBitmap; x, y : Integer);
begin
  ChangeBrushAndPen(img);
  img.Canvas.Ellipse(
    Round(x)-r,
    Round(y)-r,
    Round(x)+r,
    Round(y)+r
  );
end;

function TPatternCircle.Id : Integer;
begin
  Result := 2;
end;

function TPatternCircle.Area : Real;
begin
  Result := r*r*3.141592653589;
end;

// -------------------------- TPatternRectangle -------------------
constructor TPatternRectangle.Create(width, height : Integer);
begin
  w := width;
  h := height;
end;

procedure TPatternRectangle.Draw(img : TBitmap; x, y : Integer);
begin
  ChangeBrushAndPen(img);

  img.Canvas.Rectangle(
    Round(x)-Round(Real(w)/2.0),
    Round(y)-Round(Real(h)/2.0),
    Round(x)+Round(Real(w)/2.0),
    Round(y)+Round(Real(h)/2.0)
  );
end;

function TPatternRectangle.Id : Integer;
begin
  Result := 3;
end;

function TPatternRectangle.Area : Real;
begin
  Result := w*h;
end;


function TPatternRectangle.IsIn(x1, y1, x2, y2 : Real) : Boolean;
begin
  Result := IsIn(Round(x1), Round(y1),Round(x2),Round(y2));
end;

function TPatternRectangle.IsIn(x1, y1, x2, y2 : Integer) : Boolean;
begin
  Result :=
  (x1 >= x2 + left) and
  (y1 >= y2 + top) and
  (x1 <= x2 + right) and
  (y1 <= y2 + bottom)
  ;
end;

function TPatternRectangle.left : Integer;
begin
  Result := - Round(Real(w) / 2.0);
end;

function TPatternRectangle.right : Integer;
begin
  Result := + Round(Real(w) / 2.0);
end;
function TPatternRectangle.top : Integer;
begin
  Result := - Round(Real(h) / 2.0);
end;
function TPatternRectangle.bottom : Integer;
begin
  Result := + Round(Real(h) / 2.0);
end;

end.

